/*
 * libwebmdec: a decoder library for WebM audio/video streams
 * Copyright (c) 2014 Andrew Church <achurch@achurch.org>
 *
 * This software may be copied and redistributed under certain conditions;
 * see the file "COPYING" in the source code distribution for details.
 * NO WARRANTY is provided with this software.
 */

#ifdef DECODE_AUDIO  /* To the end of the file. */

#include "include/webmdec.h"
#include "include/internal.h"

#include <stdlib.h>
#include <string.h>

#ifdef USE_TREMOR
# include <ivorbiscodec.h>
#else
# include <vorbis/codec.h>
#endif

/*************************************************************************/
/************************ Decoder data structure *************************/
/*************************************************************************/

struct vorbis_decoder_t {
    /* Various data structures used by different parts of libvorbis. */
    vorbis_info vi;
    vorbis_comment vc;
    vorbis_block vb;
    vorbis_dsp_state vd;

    /* PCM data not yet returned to the caller. */
    float *pcm;
    int num_samples;
};

/*************************************************************************/
/**************************** Helper routines ****************************/
/*************************************************************************/

/**
 * packet:  Return a pointer to a static ogg_packet structure filled with
 * the given data.
 */
static ogg_packet *packet(const void *data, int length)
{
    static ogg_packet op = {.granulepos = -1};
    /* We de-const the pointer because ogg_packet.packet is a non-const
     * pointer type, but we assume libvorbis won't attempt to modify the
     * data. */
    op.packet = (void *)data;
    op.bytes = length;
    return &op;
}

/**
 * packet_bos:  Return a pointer to a static ogg_packet structure filled
 * with the given data and the beginning-of-stream flag set to 1.  This is
 * needed because libvorbis actually checks that the first header is in a
 * packet marked as beginning-of-stream and fails if the packet is not so
 * marked.
 */
static ogg_packet *packet_bos(const void *data, int length)
{
    static ogg_packet op = {.b_o_s = 1, .granulepos = -1};
    op.packet = (void *)data;
    op.bytes = length;
    return &op;
}

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

vorbis_decoder_t *vorbis_decoder_create(
    const unsigned char *id_header, int id_length,
    const unsigned char *setup_header, int setup_length)
{
    if (id_header == NULL
     || id_length < 7
     || memcmp(id_header, "\1vorbis", 7) != 0
     || setup_header == NULL
     || setup_length < 7
     || memcmp(setup_header, "\5vorbis", 7) != 0)
    {
        goto error_return;
    }

    vorbis_decoder_t *decoder = malloc(sizeof(*decoder));
    if (!decoder) {
        goto error_return;
    }
    decoder->pcm = NULL;
    decoder->num_samples = 0;

    vorbis_info_init(&decoder->vi);
    /* vorbis_info_init() doesn't tell us whether memory allocation failed,
     * so we have to check manually. */
    if (!decoder->vi.codec_setup) {
        goto error_free_vorbis_info;
    }

    vorbis_comment_init(&decoder->vc);

    if (vorbis_synthesis_headerin(&decoder->vi, &decoder->vc,
                                  packet_bos(id_header, id_length)) != 0) {
        goto error_free_vorbis_comment;
    }
    /* libvorbis complains if we don't give it a comment header, so fake it. */
    if (vorbis_synthesis_headerin(&decoder->vi, &decoder->vc, packet(
                                      "\3vorbis\0\0\0\0\0\0\0\0\1",16)) != 0) {
        goto error_free_vorbis_comment;
    }
    if (vorbis_synthesis_headerin(&decoder->vi, &decoder->vc,
                                  packet(setup_header, setup_length)) != 0) {
        goto error_free_vorbis_comment;
    }

    if (vorbis_synthesis_init(&decoder->vd, &decoder->vi) != 0) {
        goto error_free_vorbis_comment;
    }
    if (vorbis_block_init(&decoder->vd, &decoder->vb) != 0) {
        goto error_free_vorbis_dsp;
    }

    return decoder;

  error_free_vorbis_dsp:
    vorbis_dsp_clear(&decoder->vd);
  error_free_vorbis_comment:
    vorbis_comment_clear(&decoder->vc);
  error_free_vorbis_info:
    vorbis_info_clear(&decoder->vi);
    free(decoder);
  error_return:
    return NULL;
}

/*-----------------------------------------------------------------------*/

void vorbis_decoder_destroy(vorbis_decoder_t *decoder)
{
    free(decoder->pcm);
    vorbis_block_clear(&decoder->vb);
    vorbis_dsp_clear(&decoder->vd);
    vorbis_comment_clear(&decoder->vc);
    vorbis_info_clear(&decoder->vi);
    free(decoder);
}

/*-----------------------------------------------------------------------*/

void vorbis_decoder_reset(vorbis_decoder_t *decoder)
{
    vorbis_synthesis_restart(&decoder->vd);
}

/*-----------------------------------------------------------------------*/

int vorbis_decoder_decode(
    vorbis_decoder_t *decoder, const void *data, int length)
{
    if (vorbis_synthesis(&decoder->vb, packet(data, length)) != 0) {
        return 0;
    }
    vorbis_synthesis_blockin(&decoder->vd, &decoder->vb);

#ifdef USE_TREMOR
    ogg_int32_t **pcm;
#else
    float **pcm;
#endif
    const int new_samples = vorbis_synthesis_pcmout(&decoder->vd, &pcm);
    if (new_samples > 0) {
        const int total_samples = decoder->num_samples + new_samples;
        const int sample_size = 4 * decoder->vi.channels;
        float *new_pcm = realloc(decoder->pcm, total_samples * sample_size);
        if (!new_pcm) {
            return 0;
        }
        float *ptr = &new_pcm[decoder->num_samples * decoder->vi.channels];
        for (int i = 0; i < new_samples; i++) {
            for (int c = 0; c < decoder->vi.channels; c++, ptr++) {
#ifdef USE_TREMOR
                const ogg_int32_t sample = pcm[c][i];
                if (sample >= 32767<<9) {
                    *ptr = 1.0f;
                } else if (sample > -32767<<9) {
                    *ptr = pcm[c][i] * (1.0f/(32767<<9));
                } else {
                    *ptr = -1.0f;
                }
#else
                *ptr = pcm[c][i];
#endif
            }
        }
        decoder->pcm = new_pcm;
        decoder->num_samples = total_samples;
        vorbis_synthesis_read(&decoder->vd, new_samples);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

int vorbis_decoder_available_samples(const vorbis_decoder_t *decoder)
{
    return decoder->num_samples;
}

/*-----------------------------------------------------------------------*/

int vorbis_decoder_get_pcm(
    vorbis_decoder_t *decoder, float *buffer, int samples)
{
    if (samples > decoder->num_samples) {
        samples = decoder->num_samples;
    }
    if (samples <= 0) {
        return 0;
    }

    const int sample_size = 4 * decoder->vi.channels;
    memcpy(buffer, decoder->pcm, samples * sample_size);
    decoder->num_samples -= samples;
    if (decoder->num_samples > 0) {
        memmove(decoder->pcm, decoder->pcm + (samples * decoder->vi.channels),
                decoder->num_samples * sample_size);
        float *new_pcm =
            realloc(decoder->pcm, decoder->num_samples * sample_size);
        /* This realloc should never fail, but if it does, just leave the
         * larger buffer alone since it won't break anything. */
        if (new_pcm) {
            decoder->pcm = new_pcm;
        }
    } else {
        free(decoder->pcm);
        decoder->pcm = NULL;
    }

    return samples;
}

/*************************************************************************/
/*************************************************************************/

#endif  /* DECODE_AUDIO */
