/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/dyngl-funcs.h: Function table for dynamic loading of
 * OpenGL.
 */

/*
 * This file can be used to declare or define various sets of data for the
 * OpenGL functions used by the program.  The caller should define the
 * following preprocessor macro:
 *     DYNGL_FUNC(rettype,name,altname,params,args,required,category)
 * which will be called in the following manner:
 *     DYNGL_FUNC(void, glFunc, glFuncEXT, (GLenum a), (a), required, category)
 * for each OpenGL function to be processed.  "required" is a Boolean
 * expression evaluating to true if the function is required, 0 if it is
 * optional; for many functions, this is defined with reference to
 * variables "major" and "minor", which are expected to hold the major and
 * minor version of the OpenGL library.  (These variables do not need to be
 * defined if the "required" parameter is ignored.)  "category" is a token
 * which can be pasted onto an identifier to e.g. create a variable name
 * for use in checking the presence of that category of functions.  The
 * "altname" parameter gives an alternate name for the function (which may
 * be the same as its base name) to handle systems on which the base name
 * is not available.
 *
 * See sysdep/sdl/dyngl.c for an example of usage.
 *
 * Note that we intentionally do not protect this file against double
 * inclusion, since it may (and typically will) be included multiple times
 * with different definitions of DYNGL_FUNC.
 */

/*************************************************************************/
/*************************************************************************/

/* Shortcut macros to simplify the function list: */

/* Needed to dereference CATEGORY for token pasting. */
#define FUNC(a,b,c,d,e,f,g)  DYNGL_FUNC(a,b,c,d,e,f,g)

#define FUNC0(rettype,name) \
    FUNC(rettype, name, name, (void), (), REQUIRED_FLAG, CATEGORY)
#define FUNC1(rettype,name,type1) \
    FUNC(rettype, name, name, (type1 a), (a), REQUIRED_FLAG, CATEGORY)
#define FUNC2(rettype,name,type1,type2) \
    FUNC(rettype, name, name, (type1 a, type2 b), (a,b), REQUIRED_FLAG, \
         CATEGORY)
#define FUNC3(rettype,name,type1,type2,type3) \
    FUNC(rettype, name, name, (type1 a, type2 b, type3 c), (a,b,c), \
         REQUIRED_FLAG, CATEGORY)
#define FUNC4(rettype,name,type1,type2,type3,type4) \
    FUNC(rettype, name, name, (type1 a, type2 b, type3 c, type4 d), \
         (a,b,c,d), REQUIRED_FLAG, CATEGORY)
#define FUNC5(rettype,name,type1,type2,type3,type4, type5) \
    FUNC(rettype, name, name, (type1 a, type2 b, type3 c, type4 d, type5 e), \
         (a,b,c,d,e), REQUIRED_FLAG, CATEGORY)
#define FUNC6(rettype,name,type1,type2,type3,type4,type5,type6) \
    FUNC(rettype, name, name, (type1 a, type2 b, type3 c, type4 d, type5 e, \
                         type6 f), (a,b,c,d,e,f), REQUIRED_FLAG, CATEGORY)
#define FUNC7(rettype,name,type1,type2,type3,type4,type5,type6,type7) \
    FUNC(rettype, name, name, (type1 a, type2 b, type3 c, type4 d, type5 e, \
                               type6 f, type7 g), (a,b,c,d,e,f,g), \
         REQUIRED_FLAG, CATEGORY)
#define FUNC8(rettype,name,type1,type2,type3,type4,type5,type6,type7,type8) \
    FUNC(rettype, name, name, (type1 a, type2 b, type3 c, type4 d, type5 e, \
                               type6 f, type7 g, type8 h), (a,b,c,d,e,f,g,h), \
         REQUIRED_FLAG, CATEGORY)
#define FUNC9(rettype,name,type1,type2,type3,type4,type5,type6,type7,type8,type9) \
    FUNC(rettype, name, name, (type1 a, type2 b, type3 c, type4 d, type5 e, \
                               type6 f, type7 g, type8 h, type9 i), \
         (a,b,c,d,e,f,g,h,i), REQUIRED_FLAG, CATEGORY)

#define FUNC1_ARB(rettype,name,type1) \
    FUNC(rettype, name, name##ARB, (type1 a), (a), REQUIRED_FLAG, CATEGORY)
#define FUNC2_ARB(rettype,name,type1,type2) \
    FUNC(rettype, name, name##ARB, (type1 a, type2 b), (a,b), REQUIRED_FLAG, \
         CATEGORY)
#define FUNC4_ARB(rettype,name,type1,type2,type3,type4) \
    FUNC(rettype, name, name##ARB, (type1 a, type2 b, type3 c, type4 d), \
         (a,b,c,d), REQUIRED_FLAG, CATEGORY)

#define FUNC1_EXT(rettype,name,type1) \
    FUNC(rettype, name, name##EXT, (type1 a), (a), REQUIRED_FLAG, CATEGORY)
#define FUNC2_EXT(rettype,name,type1,type2) \
    FUNC(rettype, name, name##EXT, (type1 a, type2 b), (a,b), REQUIRED_FLAG, \
         CATEGORY)
#define FUNC4_EXT(rettype,name,type1,type2,type3,type4) \
    FUNC(rettype, name, name##EXT, (type1 a, type2 b, type3 c, type4 d), \
         (a,b,c,d), REQUIRED_FLAG, CATEGORY)
#define FUNC5_EXT(rettype,name,type1,type2,type3,type4, type5) \
    FUNC(rettype, name, name##EXT, (type1 a, type2 b, type3 c, type4 d, \
                                    type5 e), (a,b,c,d,e), REQUIRED_FLAG, \
         CATEGORY)

/*-----------------------------------------------------------------------*/

/* Base OpenGL */
#define REQUIRED_FLAG 1
#define CATEGORY /*nothing*/
FUNC2(void, glBindTexture, GLenum, GLuint)
FUNC2(void, glBlendFunc, GLenum, GLenum)
FUNC1(void, glClear, GLbitfield)
FUNC4(void, glClearColor, GLclampf, GLclampf, GLclampf, GLclampf)
FUNC1(void, glClearDepth, GLclampd)
FUNC4(void, glColorMask, GLboolean, GLboolean, GLboolean, GLboolean)
FUNC8(void, glCopyTexImage2D, GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLsizei, GLint)
FUNC8(void, glCopyTexSubImage2D, GLenum, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei)
FUNC1(void, glCullFace, GLenum)
FUNC2(void, glDeleteTextures, GLsizei, const GLuint *)
FUNC1(void, glDepthFunc, GLenum)
FUNC1(void, glDepthMask, GLboolean)
#ifdef SIL_OPENGL_ES
FUNC2(void, glDepthRangef, GLclampf, GLclampf)
#else
FUNC2(void, glDepthRange, GLclampd, GLclampd)
#endif
FUNC1(void, glDisable, GLenum)
FUNC1(void, glEnable, GLenum)
FUNC0(void, glFinish)
FUNC1(void, glFrontFace, GLenum)
FUNC2(void, glGenTextures, GLsizei, GLuint *)
#ifndef SIL_OPENGL_DISABLE_GETERROR
FUNC0(GLenum, glGetError)
#endif
FUNC2(void, glGetFloatv, GLenum, GLfloat *)
FUNC2(void, glGetIntegerv, GLenum, GLint *)
FUNC1(const GLubyte *, glGetString, GLenum)
FUNC5(void, glGetTexImage, GLenum, GLint, GLenum, GLenum, GLvoid *)
FUNC4(void, glGetTexLevelParameterfv, GLenum, GLint, GLenum, GLfloat *)
FUNC1(GLboolean, glIsEnabled, GLenum)
FUNC1(void, glLineWidth, GLfloat)
FUNC2(void, glPixelStorei, GLenum, GLint)
FUNC1(void, glPointSize, GLfloat)
FUNC2(void, glRasterPos2i, GLint, GLint)
FUNC7(void, glReadPixels, GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, GLvoid *)
FUNC4(void, glScissor, GLint, GLint, GLsizei, GLsizei)
FUNC3(void, glStencilFunc, GLenum, GLint, GLuint)
FUNC3(void, glStencilOp, GLenum, GLenum, GLenum)
FUNC9(void, glTexImage2D, GLenum, GLint, GLint, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *)
FUNC3(void, glTexParameteri, GLenum, GLenum, GLint)
FUNC9(void, glTexSubImage2D, GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *)
FUNC4(void, glViewport, GLint, GLint, GLsizei, GLsizei)

/* OpenGL 1.1 (assumed to be present) */
FUNC3(void, glDrawArrays, GLenum, GLint, GLsizei)
FUNC4(void, glDrawElements, GLenum, GLsizei, GLenum, const GLvoid *)

/* Extended blend functionality (OpenGL 1.4 or GL_EXT_blend_{color,subtract},
 * assumed to be present) */
FUNC4(void, glBlendColor, GLclampf, GLclampf, GLclampf, GLclampf)
FUNC1(void, glBlendEquation, GLenum)

/* 1.0 and 1.1 fixed function pipeline functions (omitted from OpenGL 3.0+) */
#ifndef SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS
FUNC2(void, glAlphaFunc, GLenum, GLclampf)
FUNC1(void, glBegin, GLenum)
FUNC1(void, glCallList, GLuint)
FUNC3(void, glColor3f, GLfloat, GLfloat, GLfloat)
FUNC4(void, glColor4f, GLfloat, GLfloat, GLfloat, GLfloat)
FUNC4(void, glColor4ub, GLubyte, GLubyte, GLubyte, GLubyte)
FUNC4(void, glColorPointer, GLint, GLenum, GLsizei, const GLvoid *)
FUNC2(void, glDeleteLists, GLuint, GLsizei)
FUNC1(void, glDisableClientState, GLenum)
FUNC1(void, glEnableClientState, GLenum)
FUNC0(void, glEnd)
FUNC0(void, glEndList)
FUNC2(void, glFogf, GLenum, GLfloat)
FUNC2(void, glFogfv, GLenum, const GLfloat *)
FUNC2(void, glFogi, GLenum, GLint)
FUNC1(GLuint, glGenLists, GLsizei)
FUNC0(void, glLoadIdentity)
FUNC1(void, glLoadMatrixf, const GLfloat *)
FUNC1(void, glMatrixMode, GLenum)
FUNC1(void, glMultMatrixf, const GLfloat *)
FUNC2(void, glNewList, GLuint, GLenum)
FUNC6(void, glOrtho, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble)
FUNC0(void, glPopAttrib)
FUNC0(void, glPopClientAttrib)
FUNC0(void, glPopMatrix)
FUNC1(void, glPushAttrib, GLbitfield)
FUNC1(void, glPushClientAttrib, GLbitfield)
FUNC0(void, glPushMatrix)
FUNC4(void, glRotatef, GLfloat, GLfloat, GLfloat, GLfloat)
FUNC3(void, glScalef, GLfloat, GLfloat, GLfloat)
FUNC2(void, glTexCoord2f, GLfloat, GLfloat)
FUNC4(void, glTexCoordPointer, GLint, GLenum, GLsizei, const GLvoid *)
FUNC3(void, glTranslatef, GLfloat, GLfloat, GLfloat)
FUNC2(void, glVertex2i, GLint, GLint)
FUNC2(void, glVertex2f, GLfloat, GLfloat)
FUNC3(void, glVertex3f, GLfloat, GLfloat, GLfloat)
FUNC4(void, glVertexPointer, GLint, GLenum, GLsizei, const GLvoid *)
#endif  // !SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS

/*----------------------------------*/

/* Compressed textures (OpenGL 1.3, ARB_texture_compression) */
#undef REQUIRED_FLAG
#define REQUIRED_FLAG (major >= 2 || (major == 1 && minor >= 3))
#undef CATEGORY
#define CATEGORY compressed_textures
FUNC8(void, glCompressedTexImage2D, GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *)

/* Vertex buffers (OpenGL 1.5, ARB_vertex_buffer_object) */
#undef REQUIRED_FLAG
#define REQUIRED_FLAG (major >= 2 || (major == 1 && minor >= 5))
#undef CATEGORY
#define CATEGORY vertex_buffers
FUNC2_ARB(void, glBindBuffer, GLenum, GLuint)
FUNC4_ARB(void, glBufferData, GLenum, GLsizeiptr, const GLvoid *, GLenum)
FUNC4_ARB(void, glBufferSubData, GLenum, GLintptr, GLsizeiptr, const GLvoid *)
FUNC2_ARB(void, glDeleteBuffers, GLsizei, const GLuint *)
FUNC2_ARB(void, glGenBuffers, GLsizei, GLuint *)
FUNC4_ARB(void, glGetBufferSubData, GLenum, GLintptr, GLsizeiptr, GLvoid *)
FUNC2_ARB(void *, glMapBuffer, GLenum, GLenum)
FUNC1_ARB(GLboolean, glUnmapBuffer, GLenum)

/* Shaders (OpenGL 2.0; the earlier ARBs had a different calling format, so
 * we don't allow fallback) */
#undef REQUIRED_FLAG
#define REQUIRED_FLAG (major >= 2)
#undef CATEGORY
#define CATEGORY shaders
FUNC2(void, glAttachShader, GLuint, GLuint)
FUNC3(void, glBindAttribLocation, GLuint, GLuint, const GLchar *)
FUNC1(void, glCompileShader, GLuint)
FUNC0(GLuint, glCreateProgram)
FUNC1(GLuint, glCreateShader, GLenum)
FUNC1(void, glDeleteProgram, GLuint)
FUNC1(void, glDeleteShader, GLuint)
FUNC2(void, glDetachShader, GLuint, GLuint)
FUNC1(void, glDisableVertexAttribArray, GLuint)
FUNC1(void, glEnableVertexAttribArray, GLuint)
FUNC3(void, glGetProgramiv, GLuint, GLenum, GLint *)
FUNC4(void, glGetProgramInfoLog, GLuint, GLsizei, GLsizei *, GLchar *)
FUNC3(void, glGetShaderiv, GLuint, GLenum, GLint *)
FUNC4(void, glGetShaderInfoLog, GLuint, GLsizei, GLsizei *, GLchar *)
FUNC4(void, glGetShaderSource, GLuint, GLsizei, GLsizei *, GLchar *)
FUNC2(GLint, glGetUniformLocation, GLuint, const GLchar *)
FUNC1(GLboolean, glIsProgram, GLuint)
FUNC1(GLboolean, glIsShader, GLuint)
FUNC1(void, glLinkProgram, GLuint)
FUNC4(void, glShaderSource, GLuint, GLsizei, const GLchar **, const GLint *)
FUNC2(void, glUniform1f, GLint, GLfloat)
FUNC3(void, glUniform1fv, GLint, GLsizei, const GLfloat *)
FUNC2(void, glUniform1i, GLint, GLint)
FUNC3(void, glUniform1iv, GLint, GLsizei, const GLint *)
FUNC3(void, glUniform2f, GLint, GLfloat, GLfloat)
FUNC3(void, glUniform2fv, GLint, GLsizei, const GLfloat *)
FUNC3(void, glUniform2i, GLint, GLint, GLint)
FUNC3(void, glUniform2iv, GLint, GLsizei, const GLint *)
FUNC4(void, glUniform3f, GLint, GLfloat, GLfloat, GLfloat)
FUNC3(void, glUniform3fv, GLint, GLsizei, const GLfloat *)
FUNC4(void, glUniform3i, GLint, GLint, GLint, GLint)
FUNC3(void, glUniform3iv, GLint, GLsizei, const GLint *)
FUNC5(void, glUniform4f, GLint, GLfloat, GLfloat, GLfloat, GLfloat)
FUNC3(void, glUniform4fv, GLint, GLsizei, const GLfloat *)
FUNC5(void, glUniform4i, GLint, GLint, GLint, GLint, GLint)
FUNC3(void, glUniform4iv, GLint, GLsizei, const GLint *)
FUNC4(void, glUniformMatrix2fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC4(void, glUniformMatrix2x3fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC4(void, glUniformMatrix2x4fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC4(void, glUniformMatrix3fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC4(void, glUniformMatrix3x2fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC4(void, glUniformMatrix3x4fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC4(void, glUniformMatrix4fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC4(void, glUniformMatrix4x2fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC4(void, glUniformMatrix4x3fv, GLint, GLsizei, GLboolean, const GLfloat *)
FUNC1(void, glUseProgram, GLuint)
FUNC1(void, glValidateProgram, GLuint)
FUNC6(void, glVertexAttribPointer, GLuint, GLint, GLenum, GLboolean, GLsizei, const GLvoid *)

/* Framebuffers (OpenGL 3.0, EXT_framebuffer_object) */
#undef REQUIRED_FLAG
#define REQUIRED_FLAG (major >= 3)
#undef CATEGORY
#define CATEGORY framebuffers
FUNC2_EXT(void, glBindFramebuffer, GLenum, GLuint)
FUNC2_EXT(void, glBindRenderbuffer, GLenum, GLuint)
FUNC1_EXT(GLenum, glCheckFramebufferStatus, GLenum)
FUNC2_EXT(void, glDeleteFramebuffers, GLsizei, const GLuint *)
FUNC2_EXT(void, glDeleteRenderbuffers, GLsizei, const GLuint *)
FUNC4_EXT(void, glFramebufferRenderbuffer, GLenum, GLenum, GLenum, GLuint)
FUNC5_EXT(void, glFramebufferTexture2D, GLenum, GLenum, GLenum, GLuint, GLint)
FUNC2_EXT(void, glGenFramebuffers, GLsizei, GLuint *)
FUNC2_EXT(void, glGenRenderbuffers, GLsizei, GLuint *)
/* Despite being totally unrelated to framebuffers, glGenerateMipmap() is
 * myseriously part of the EXT_framebuffer_object extension. */
FUNC1_EXT(void, glGenerateMipmap, GLenum)
FUNC4_EXT(void, glGetFramebufferAttachmentParameteriv, GLenum, GLenum, GLenum, GLint *)
FUNC4_EXT(void, glRenderbufferStorage, GLenum, GLenum, GLsizei, GLsizei)

/* OpenGL 3.0 miscellaneous */
#undef REQUIRED_FLAG
#define REQUIRED_FLAG (major >= 3)
#undef CATEGORY
#define CATEGORY opengl3
FUNC3(void, glBindFragDataLocation, GLuint, GLuint, const GLchar *)
FUNC1(void, glBindVertexArray, GLuint)
FUNC2(void, glDeleteVertexArrays, GLsizei, const GLuint *)
FUNC2(void, glGenVertexArrays, GLsizei, GLuint *)
FUNC2(const GLubyte *, glGetStringi, GLenum, GLuint)


/*-----------------------------------------------------------------------*/

#undef FUNC
#undef FUNC0
#undef FUNC1
#undef FUNC2
#undef FUNC3
#undef FUNC4
#undef FUNC5
#undef FUNC6
#undef FUNC7
#undef FUNC8
#undef FUNC9
#undef FUNC1_ARB
#undef FUNC2_ARB
#undef FUNC4_ARB
#undef FUNC1_EXT
#undef FUNC2_EXT
#undef FUNC4_EXT
#undef FUNC5_EXT
#undef REQUIRED_FLAG
#undef CATEGORY

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
