/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/endian.c: Tests for endian-conversion functions.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/endian.h"
#include "src/test/base.h"

/*************************************************************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_endian)

/*-----------------------------------------------------------------------*/

TEST(test_endian_detection)
{
    union {
        char c[4];
        uint32_t i;
    } buf = {.c = {0x12, 0x34, 0x56, 0x78}};
    CHECK_INTEQUAL(buf.i, is_little_endian() ? 0x78563412 : 0x12345678);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_be_s16)
{
    const int16_t s16_native  = -0x1234;
    const int16_t s16_be      = (is_little_endian() ? -0x3313 : -0x1234);

    /* We explicitly sign-extend values to the next larger integer size
     * in order to check the return value's signedness. */
    CHECK_INTEQUAL((int32_t)be_to_s16(s16_be), (int32_t)s16_native);
    CHECK_INTEQUAL((int32_t)s16_to_be(s16_native), (int32_t)s16_be);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_be_u16)
{
    const uint16_t u16_native = 0xCDEF;
    const uint16_t u16_be     = (is_little_endian() ? 0xEFCD : 0xCDEF);

    CHECK_INTEQUAL((int32_t)be_to_u16(u16_be), (int32_t)u16_native);
    CHECK_INTEQUAL((int32_t)u16_to_be(u16_native), (int32_t)u16_be);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_be_s32)
{
    const int32_t s32_native  = -0x12345678;
    const int32_t s32_be      = (is_little_endian() ? -0x77563413:-0x12345678);

    CHECK_INTEQUAL((int64_t)be_to_s32(s32_be), (int64_t)s32_native);
    CHECK_INTEQUAL((int64_t)s32_to_be(s32_native), (int64_t)s32_be);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_be_u32)
{
    const uint32_t u32_native = 0x89ABCDEF;
    const uint32_t u32_be     = (is_little_endian() ? 0xEFCDAB89 : 0x89ABCDEF);

    CHECK_INTEQUAL((int64_t)be_to_u32(u32_be), (int64_t)u32_native);
    CHECK_INTEQUAL((int64_t)u32_to_be(u32_native), (int64_t)u32_be);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_be_float)
{
    const float float_native = 1.0f;
    const union {uint32_t i; float f;} float_be =
        {.i = (is_little_endian() ? 0x0000803F : 0x3F800000)};
    const union {float f; uint32_t i;} float_test =
        {.f = float_to_be(float_native)};

    CHECK_INTEQUAL(float_test.i, float_be.i);
    CHECK_FLOATEQUAL(be_to_float(float_be.f), float_native);

    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
