/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/thread.c: Tests for the thread management functions.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/math.h"
#include "src/test/base.h"
#include "src/thread.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Number of iterations to spin on thread_yield() while waiting for a
 * subthread.  Used in both performance tests and test_thread_is_running(). */
#define SPIN_COUNT  100000

#ifdef SIL_TEST_THREAD_PERFORMANCE

/* Number of times to run the thread for performance measurement. */
# define THREAD_RUNS  5

/* Positive and negative priority values to use for performance measurement.
 * These should be of sufficient magnitude to ensure that differences will
 * be properly measured on all systems which support thread priorities. */
# define POSITIVE_PRIORITY  +10
# define NEGATIVE_PRIORITY  -10

/* Tolerance for comparing performance values.  For equality tests, the
 * test succeeds if either value is within this fraction of the other.
 * For inequality tests, a > b succeeds if a > b*(1-TOLERANCE) or
 * a*(1+TOLERANCE) > b. */
# define TOLERANCE  0.25

#endif  // SIL_TEST_THREAD_PERFORMANCE

/*-----------------------------------------------------------------------*/

/* Data structure passed to counter_thread(). */
typedef struct CounterData CounterData;
struct CounterData {
    uint32_t counter;
    uint8_t stop;
};

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * run_counter_thread:  Run the counter thread for a while and return its
 * result.
 *
 * A successful run is defined as successfully creating the thread and
 * getting a return value from thread_wait() equal to the counter value in
 * the parameter block.  In particular, a return value of zero (meaning no
 * iterations were executed inside the thread) is still considered successful.
 *
 * [Parameters]
 *     use_priority: If true, call sys_create_thread_with_priority(); if
 *         false, call sys_create_thread().
 *     priority: Priority value to use (when use_priority != 0).
 *     num_runs: Number of times to run the thread.
 *     spin_count: Number of times to spin on the thread_yield() before
 *         terminating the thread; if zero, spins until the thread's
 *         counter variable becomes nonzero.
 *     result_ret: Pointer to variable to receive sum of thread return values.
 * [Return value]
 *     True if the thread ran successfully, false if not.
 */
static int run_counter_thread(int use_priority, int priority, int num_runs,
                              int spin_count, int *result_ret);

/**
 * counter_thread:  Thread routine that counts endlessly until its
 * parameter block's stop flag is set, then returns the final count.
 *
 * [Parameters]
 *     param: Pointer to parameter block (CounterData).
 * [Return value]
 *     Number of iterations executed (== CounterData.counter).
 */
static int counter_thread(void *param);

/**
 * large_stack_thread:  Thread routine that allocates a 124*1024-byte array
 * on the stack, writes values to the entire array buffer, and performs
 * arithmetic on those values to ensure that the entire array is accessible.
 *
 * [Parameters]
 *     param_unused: Unused.
 * [Return value]
 *     503824896
 */
static int large_stack_thread(void *param_unused);

/**
 * waiter_thread:  Thread routine that waits for the value pointed to by
 * its parameter to go nonzero, calls thread_wait() on that value, sets
 * the value pointed to by its parameter to zero, and returns the result
 * of thread_wait().
 *
 * [Parameters]
 *     param: Pointer to int variable (see function description).
 * [Return value]
 *     Result of thread_wait().
 */
static int waiter_thread(void *param);

/**
 * waiter2_thread:  Alternate version of waiter_thread() which calls
 * thread_wait2() instead of thread_wait().
 *
 * [Parameters]
 *     param: Pointer to int variable (see waiter_thread() function
 *         description).
 * [Return value]
 *     Result of thread_wait2().
 */
static int waiter2_thread(void *param);

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_thread)

/*-----------------------------------------------------------------------*/

TEST(test_thread_create)
{
    int result;
    if (!run_counter_thread(0, 0, 1, 0, &result)) {
        FAIL("Thread run failed");
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_create_memory_failure)
{
#ifndef SIL_PLATFORM_PSP  // PSP doesn't need to allocate memory for threads.
    int result;
    CHECK_MEMORY_FAILURES(run_counter_thread(0, 0, 1, 0, &result));
#endif

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_create_with_priority)
{
    int result;
    if (!run_counter_thread(1, 0, 1, 0, &result)) {
        FAIL("Thread run failed");
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_create_with_stack_size)
{
    int thread;

    CHECK_TRUE(thread = thread_create_with_stack_size(
                   0, 131072, large_stack_thread, NULL));
    CHECK_INTEQUAL(thread_wait(thread), 503824896);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_create_invalid)
{
    CHECK_FALSE(thread_create(NULL, NULL));
    CHECK_FALSE(thread_create_with_priority(0, NULL, NULL));
    CHECK_FALSE(thread_create_with_stack_size(0, 0, NULL, NULL));
    CHECK_FALSE(thread_create_with_stack_size(0, -1, counter_thread, NULL));

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_create_forced_failure)
{
    TEST_thread_fail_create(1);

    CHECK_FALSE(thread_create(counter_thread, NULL));
    CHECK_FALSE(thread_create_with_priority(0, counter_thread, NULL));
    CHECK_FALSE(thread_create_with_stack_size(0, 0, counter_thread, NULL));

    TEST_thread_fail_create(0);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_is_running)
{
    int thread;
    CounterData counter_data;

    counter_data.counter = 0;
    counter_data.stop = 0;
    thread = thread_create(counter_thread, &counter_data);
    if (!thread) {
        FAIL("Failed to create counter thread");
    }
    if (!thread_is_running(thread)) {
        FAIL("thread_is_running() reported running thread as stopped");
    }
    counter_data.stop = 1;
    for (int try = 0; thread_is_running(thread) && try < SPIN_COUNT; try++) {
        thread_yield();
    }
    if (thread_is_running(thread)) {
        thread_wait(thread);
        FAIL("thread_is_running() reported stopped thread as running (or"
             " system is too loaded for this test)");
    }
    thread_wait(thread);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_wait2)
{
    int thread;
    int result = 0;

    CHECK_TRUE(thread = thread_create_with_stack_size(
                   0, 131072, large_stack_thread, NULL));
    CHECK_TRUE(thread_wait2(thread, &result));
    CHECK_INTEQUAL(result, 503824896);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_wait_self)
{
    int thread;
    int param = 0;

    CHECK_TRUE(thread = thread_create(waiter_thread, &param));
    param = thread;
    do {
        thread_yield();
    } while (param);
    CHECK_INTEQUAL(thread_wait(thread), 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_wait2_self)
{
    int thread;
    int param = 0;

    CHECK_TRUE(thread = thread_create(waiter2_thread, &param));
    param = thread;
    do {
        thread_yield();
    } while (param);
    CHECK_INTEQUAL(thread_wait(thread), 0);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_thread_invalid)
{
    int thread, result;

    CHECK_FALSE(thread_create(NULL, NULL));
    CHECK_FALSE(thread_create_with_priority(0, NULL, NULL));
    CHECK_FALSE(thread_is_running(0));
    CHECK_FALSE(thread_is_running(INT_MAX));
    CHECK_FALSE(thread_wait(0));
    CHECK_FALSE(thread_wait(INT_MAX));
    CHECK_FALSE(thread_wait2(0, &result));
    CHECK_FALSE(thread_wait2(INT_MAX, &result));

    CHECK_TRUE(thread = thread_create_with_stack_size(
                   0, 131072, large_stack_thread, NULL));
    CHECK_FALSE(thread_wait2(thread, NULL));
    CHECK_INTEQUAL(thread_wait(thread), 503824896);
    CHECK_FALSE(thread_wait(thread));
    CHECK_FALSE(thread_wait2(thread, &result));

    return 1;
}

/*-----------------------------------------------------------------------*/

#ifdef SIL_TEST_THREAD_PERFORMANCE

TEST(test_thread_priority)
{
    /* Check that a thread run explicitly at priority zero performs the
     * same as a thread run without an explicit priority. */
    int sum_nopri;
    if (!run_counter_thread(0, 0, THREAD_RUNS, SPIN_COUNT, &sum_nopri)) {
        FAIL("Thread run failed");
    }
    int sum_pri0;
    if (!run_counter_thread(1, 0, THREAD_RUNS, SPIN_COUNT, &sum_pri0)) {
        FAIL("Thread run failed");
    }
    if (fabs(sum_pri0 - sum_nopri) > sum_nopri * TOLERANCE
     && fabs(sum_pri0 - sum_nopri) > sum_pri0 * TOLERANCE) {
        FAIL("Non-priority thread and priority 0 thread perform differently"
             " (non-priority: %d, priority 0: %d).  NOTE: This test may be"
             " affected by system load.", sum_nopri, sum_pri0);
    }

    /* Check that threads at positive and negative priorities run for
     * respectively no less and no more time than a priority-zero thread. */
    int sum_pos;
    if (!run_counter_thread(1, POSITIVE_PRIORITY, THREAD_RUNS, SPIN_COUNT,
                            &sum_pos)) {
        FAIL("Thread run failed");
    }
    if (sum_pos < sum_pri0*(1-TOLERANCE) && sum_pos*(1+TOLERANCE) < sum_pri0) {
        FAIL("Positive-priority thread runs shorter then priority 0 thread"
             " (priority %d: %d, priority 0: %d).  NOTE: This test may be"
             " affected by system load.", POSITIVE_PRIORITY, sum_pos,
             sum_pri0);
    }
    int sum_neg;
    if (!run_counter_thread(1, NEGATIVE_PRIORITY, THREAD_RUNS, SPIN_COUNT,
                            &sum_neg)) {
        FAIL("Thread run failed");
    }
    if (sum_pri0 < sum_neg*(1-TOLERANCE) && sum_pri0*(1+TOLERANCE) < sum_neg) {
        FAIL("Negative-priority thread runs longer then priority 0 thread"
             " (priority %d: %d, priority 0: %d).  NOTE: This test may be"
             " affected by system load.", NEGATIVE_PRIORITY, sum_neg,
             sum_pri0);
    }

    return 1;
}

#endif  // SIL_TEST_THREAD_PERFORMANCE

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static int run_counter_thread(int use_priority, int priority, int num_runs,
                              int spin_count, int *result_ret)
{
    PRECOND(result_ret != NULL, return 0);

    *result_ret = 0;

    for (int run = 1; run <= num_runs; run++) {
        CounterData counter_data;
        int thread;
        int result;

        counter_data.counter = 0;
        counter_data.stop = 0;
        if (use_priority) {
            thread = thread_create_with_priority(
                priority, counter_thread, &counter_data);
        } else {
            thread = thread_create(counter_thread, &counter_data);
        }
        if (!thread) {
            DLOG("Failed to create counter thread (run %d/%d)", run, num_runs);
            return 0;
        }

        if (spin_count > 0) {
            for (int spin = 0; spin < spin_count; spin++) {
                thread_yield();
            }
        } else {
            while (counter_data.counter == 0) {
                thread_yield();
            }
        }

        counter_data.stop = 1;
        result = thread_wait(thread);
        if (result != (int)counter_data.counter) {
            DLOG("Wrong return value from thread_wait (run %d/%d: got %d,"
                 " should be %d", run, num_runs, result,
                 (int)counter_data.counter);
            return 0;
        }

        *result_ret += result;
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

static int counter_thread(void *param)
{
    CounterData *data = (CounterData *)param;
    while (!data->stop) {
        data->counter++;
#ifdef SIL_TEST_THREAD_PERFORMANCE
        BARRIER();
#else
        thread_yield();
#endif
    }
    return (int)data->counter;
}

/*-----------------------------------------------------------------------*/

static int large_stack_thread(UNUSED void *param_unused)
{
    /* Declare the array volatile to ensure that the memory accesses
     * actually occur, since otherwise the compiler could theoretically
     * optimize this entire routine down to a single return statement. */
    volatile uint32_t array[(124*1024)/4];

    for (int i = 0; i < lenof(array); i++) {
        array[i] = i;
    }
    int sum = 0;
    for (int i = 0; i < lenof(array); i++) {
        sum += array[i];
    }
    return sum;
}

/*-----------------------------------------------------------------------*/

static int waiter_thread(void *param)
{
    int *thread_ptr = (int *)param;
    while (!*thread_ptr) {
        thread_yield();
    }
    const int result = thread_wait(*thread_ptr);
    *thread_ptr = 0;
    BARRIER();
    return result;
}

/*-----------------------------------------------------------------------*/

static int waiter2_thread(void *param)
{
    int *thread_ptr = (int *)param;
    while (!*thread_ptr) {
        thread_yield();
    }
    int thread_result;
    const int result = thread_wait2(*thread_ptr, &thread_result);
    *thread_ptr = 0;
    BARRIER();
    return result;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
