/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/utility/yuv2rgb.c: Tests for YUV->RGB conversion.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/test/base.h"
#include "src/utility/yuv2rgb.h"

#if defined(SIL_ARCH_X86) || defined(SIL_PLATFORM_PSP)
# include "src/memory.h"
#endif

#ifdef SIL_PLATFORM_PSP
# include "src/sysdep/psp/internal.h"  // For syscall declarations.
# include "src/sysdep/psp/thread.h"    // For psp_threads_{lock,unlock}().
#endif

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_utility_yuv2rgb)

/*-----------------------------------------------------------------------*/

TEST(test_Y)
{
    static const uint8_t Y[4] = {83,83,83,83};
    static const uint8_t U[1] = {128};
    static const uint8_t V[1] = {128};
    const int grey = 78;  // 78.016 => 78 for both truncation and rounding

    /* For these tests, we want to test the base C implementation, so we use
     * an unaligned destination buffer because that fails the conditions for
     * all assembly routines. */
    uint8_t rgb_base[1+4*4];
    uint8_t *rgb = rgb_base + 1;

    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){2,1,1}, rgb, 2, 2, 2);
    for (int i = 0; i < 4; i++) {
        CHECK_PIXEL(&rgb[i*4], grey,grey,grey,255, i%2, i/2);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_Y_multiple)
{
    static const uint8_t Y[4] = {83,86,89,92};
    static const uint8_t U[1] = {128};
    static const uint8_t V[1] = {128};
    static const uint8_t grey[4] = {78,82,85,88};

    uint8_t rgb_base[1+4*4];
    uint8_t *rgb = rgb_base + 1;
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){2,1,1}, rgb, 2, 2, 2);
    for (int i = 0; i < 4; i++) {
        CHECK_PIXEL(&rgb[i*4], grey[i],grey[i],grey[i],255, i%2, i/2);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_rounding)
{
    static const uint8_t Y[4] = {86,86,86,86};
    static const uint8_t U[1] = {128};
    static const uint8_t V[1] = {128};
    const int grey = 82;  // 81.510 => should round up to 82

    uint8_t rgb_base[1+4*4];
    uint8_t *rgb = rgb_base + 1;
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){2,1,1}, rgb, 2, 2, 2);
    for (int i = 0; i < 4; i++) {
        CHECK_PIXEL(&rgb[i*4], grey,grey,grey,255, i%2, i/2);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_bound)
{
    static const uint8_t Y[4] = {1,1,254,254};
    static const uint8_t U[1] = {128};
    static const uint8_t V[1] = {128};

    uint8_t rgb_base[1+4*4];
    uint8_t *rgb = rgb_base + 1;
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){2,1,1}, rgb, 2, 2, 2);
    for (int i = 0; i < 4; i++) {
        const int grey = (i < 2) ? 0 : 255;
        CHECK_PIXEL(&rgb[i*4], grey,grey,grey,255, i%2, i/2);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_U)
{
    static const uint8_t Y[4] = {86,86,86,86};
    static const uint8_t U[1] = {64};
    static const uint8_t V[1] = {128};

    uint8_t rgb_base[1+4*4];
    uint8_t *rgb = rgb_base + 1;
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){2,1,1}, rgb, 2, 2, 2);
    for (int i = 0; i < 4; i++) {
        CHECK_PIXEL(&rgb[i*4], 82,107,0,255, i%2, i/2);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_V)
{
    static const uint8_t Y[4] = {86,86,86,86};
    static const uint8_t U[1] = {128};
    static const uint8_t V[1] = {64};

    uint8_t rgb_base[1+4*4];
    uint8_t *rgb = rgb_base + 1;
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){2,1,1}, rgb, 2, 2, 2);
    for (int i = 0; i < 4; i++) {
        CHECK_PIXEL(&rgb[i*4], 0,134,82,255, i%2, i/2);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_UV_multiple)
{
    static const uint8_t Y[16] =
        {95,95,95,95,95,95,95,95,95,95,95,95,95,95,95,95};
    static const uint8_t U[4] = {80,112,144,176};
    static const uint8_t V[4] = {160,128,96,64};
    static const uint8_t expect[16*3] = {
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        143, 85,  0, 130, 88, 11, 105, 95, 44,  92, 98, 60,
        118, 92, 27, 105, 95, 44,  79,102, 76,  66,105, 92,
         66,105, 92,  54,108,108,  28,115,140,  15,118,157,
         41,112,124,  28,115,140,   3,122,173,   0,125,189,
#else
        143, 85,  0, 143, 85,  0,  92, 98, 60,  92, 98, 60,
        143, 85,  0, 143, 85,  0,  92, 98, 60,  92, 98, 60,
         41,112,124,  41,112,124,   0,125,189,   0,125,189,
         41,112,124,  41,112,124,   0,125,189,   0,125,189,
#endif
    };

    uint8_t rgb_base[1+16*4];
    uint8_t *rgb = rgb_base + 1;
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){4,2,2}, rgb, 4, 4, 4);
    for (int i = 0; i < 16; i++) {
        CHECK_PIXEL(&rgb[i*4],
                    expect[i*3+0], expect[i*3+1], expect[i*3+2], 255,
                    i%4, i/4);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

#ifdef SIL_ARCH_X86

TEST(test_asm_sse2)
{
    static const ALIGNED(16) uint8_t Y[32] =
        {50,60,70,80,90,100,110,120,130,140,150,160,170,180,190,200,
         55,65,75,85,95,105,115,125,135,145,155,165,175,185,195,205};
    /* These two are deliberately misaligned (the first byte is ignored)
     * to verify that alignment of this data is not required. */
    static const ALIGNED(8) uint8_t U[1+8] =
        {0,  8, 40, 72,104,136,168,200,232};
    static const ALIGNED(8) uint8_t V[1+8] =
        {0,248,216,184,152,120, 88, 56, 24};

    /* Note that several values here are off by 1 from what the C routine
     * produces (see test_asm_sse2_conditions() below) due to reduced
     * precision in the SSE2 routine causing different rounding.  These
     * are marked with comments. */
    static const uint8_t expect[32*3] = {
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        231,  0,  0, 230,  4,  0, 216, 22,  0, 202, 41,  0,
        188, 59,  0, 174, 77, // Rounding: 78->77
                               1, 161, 96, 45, 147,114, 89,
        133,133,133, 119,151,177, 105,169, // Rounding: 170->169
                                          221,  91,188,255,
         77,206,255,  63,225,255,  49,243,255,  48,255,255,
        237,  0,  0, 236, 10,  0, 222, 28,  0, 208, 47,  0,
        194, 65,  0, 180, 83,  7, 166,102, 51, 152,120, 95,
        139,139,139, 125,157,182, 111,175,226,  97,194,255,
         83,212,255,  69,230,255,  55,249,255,  54,255,255,
#else
        231,  0,  0, 243,  1,  0, 203, 26,  0, 215, 37,  0,
        176, 62, // Rounding: 63->62
                  0, 187, 74,  0, 148, 99, 61, 159,111, 73,
        120,136,149, 132,148,161,  92,173,237, 104,184, // Rounding: 185->184
                                                       248,
         64,210,255,  76,221,255,  37,246,255,  48,255,255,
        237,  0,  0, 249,  6, // Rounding: 7->6
                               0, 209, 32,  0, 221, 43,  0,
        181, 68,  0, 193, 80,  0, 154,105, 67, 165,117, 78, // Rounding: 79->78
        126,142,155, 137,154,166,  98,179,243, 110,190,254,
         70,215,255,  82,227,255,  42,252,255,  54,255,255,
#endif
    };

    uint8_t rgb_base[32*4+15];
    uint8_t *rgb = (uint8_t *)align_up((uintptr_t)rgb_base, 16);
    yuv2rgb((const uint8_t *[3]){Y,U+1,V+1}, (const int[3]){16,8,8}, rgb, 16,
            16, 2);
    for (int i = 0; i < 32; i++) {
        CHECK_PIXEL(&rgb[i*4],
                    expect[i*3+0], expect[i*3+1], expect[i*3+2], 255,
                    i%16, i/16);
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_asm_sse2_conditions)
{
    static const ALIGNED(16) uint8_t Y[32+33+32] =
        {50,60,70,80,90,100,110,120,130,140,150,160,170,180,190,200,
         55,65,75,85,95,105,115,125,135,145,155,165,175,185,195,205,
         50,60,70,80,90,100,110,120,130,140,150,160,170,180,190,200,
         0,
         55,65,75,85,95,105,115,125,135,145,155,165,175,185,195,205,
         50,60,70,80,90,100,110,120,130,140,150,160,170,180,190,200,
         55,65,75,85,95,105,115,125,135,145,155,165,175,185,195,205};
    static const ALIGNED(8) uint8_t U[1+8+1] =
        {0,  8, 40, 72,104,136,168,200,232,232};
    static const ALIGNED(8) uint8_t V[1+8+1] =
        {0,248,216,184,152,120, 88, 56, 24, 24};
    static const uint8_t expect[32*3] = {
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        231,  0,  0, 230,  4,  0, 216, 22,  0, 202, 41,  0,
        188, 59,  0, 174, 78,  1, 161, 96, 45, 147,114, 89,
        133,133,133, 119,151,177, 105,170,221,  91,188,255,
         77,206,255,  63,225,255,  49,243,255,  48,255,255,
        237,  0,  0, 236, 10,  0, 222, 28,  0, 208, 47,  0,
        194, 65,  0, 180, 83,  7, 166,102, 51, 152,120, 95,
        139,139,139, 125,157,182, 111,175,226,  97,194,255,
         83,212,255,  69,230,255,  55,249,255,  54,255,255,
#else
        231,  0,  0, 243,  1,  0, 203, 26,  0, 215, 37,  0,
        176, 63,  0, 187, 74,  0, 148, 99, 61, 159,111, 73,
        120,136,149, 132,148,161,  92,173,237, 104,185,248,
         64,210,255,  76,221,255,  37,246,255,  48,255,255,
        237,  0,  0, 249,  7,  0, 209, 32,  0, 221, 43,  0,
        181, 68,  0, 193, 80,  0, 154,105, 67, 165,117, 79,
        126,142,155, 137,154,166,  98,179,243, 110,190,254,
         70,215,255,  82,227,255,  42,252,255,  54,255,255,
#endif
    };

    uint8_t rgb_base[(32*4+4)+15];
    uint8_t *rgb = (uint8_t *)align_up((uintptr_t)rgb_base, 16);

    /* We fail each condition for the SSE2 routine individually to ensure
     * that all are checked.  A missed check will cause the program to
     * crash due to a misaligned SSE register store. */
    struct {
        const uint8_t *yuv[3];
        int yuv_stride[3];
        uint8_t *rgb;
        int rgb_stride;
        int width;
    } argsets[] = {
        {{Y+65,U+1,V+1}, {16,8,8}, rgb, 16, 16},
        {{Y+32,U+1,V+1}, {17,8,8}, rgb, 16, 16},
        {{Y,U+1,V+1}, {16,8,8}, rgb+1, 16, 16},
        {{Y,U+1,V+1}, {16,8,8}, rgb, 17, 16},
        {{Y,U+1,V+1}, {16,8,8}, rgb, 16, 18},
    };

    for (int set = 0; set < lenof(argsets); set++) {
        DLOG("Testing set %d", set);
        yuv2rgb(argsets[set].yuv, argsets[set].yuv_stride,
                argsets[set].rgb, argsets[set].rgb_stride,
                argsets[set].width, 2);
        for (int y = 0; y < 2; y++) {
            uint8_t *line = argsets[set].rgb + y*(argsets[set].rgb_stride*4);
            for (int x = 0; x < 16; x++) {
                CHECK_PIXEL(&line[x*4],
                            expect[(y*16+x)*3+0],
                            expect[(y*16+x)*3+1],
                            expect[(y*16+x)*3+2],
                            255,
                            x, y);
            }
        }
    }

    return 1;
}

#endif  // SIL_ARCH_X86

/*-----------------------------------------------------------------------*/

#ifdef SIL_ARCH_MIPS

TEST(test_asm_mips)
{
    static const uint8_t Y[1+32] =
        {0,50,60,70,80,90,100,110,120,130,140,150,160,170,180,190,200,
           55,65,75,85,95,105,115,125,135,145,155,165,175,185,195,205};
    static const ALIGNED(8) uint8_t U[1+8] =
        {0,  8, 40, 72,104,136,168,200,232};
    static const ALIGNED(8) uint8_t V[1+8] =
        {0,248,216,184,152,120, 88, 56, 24};
    static const uint8_t expect[32*3] = {
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        231,  0,  0, 230,  4,  0, 216, 22,  0, 202, 41,  0,
        188, 59,  0, 174, 78,  1, 161, 96, 45, 147,114, 89,
        133,133,133, 119,151,177, 105,170,221,  91,188,255,
         77,206,255,  63,225,255,  49,243,255,  48,255,255,
        237,  0,  0, 236, 10,  0, 222, 28,  0, 208, 47,  0,
        194, 65,  0, 180, 83,  7, 166,102, 51, 152,120, 95,
        139,139,139, 125,157,182, 111,175,226,  97,194,255,
         83,212,255,  69,230,255,  55,249,255,  54,255,255,
#else
        231,  0,  0, 243,  1,  0, 203, 26,  0, 215, 37,  0,
        176, 63,  0, 187, 74,  0, 148, 99, 61, 159,111, 73,
        120,136,149, 132,148,161,  92,173,237, 104,185,248,
         64,210,255,  76,221,255,  37,246,255,  48,255,255,
        237,  0,  0, 249,  7,  0, 209, 32,  0, 221, 43,  0,
        181, 68,  0, 193, 80,  0, 154,105, 67, 165,117, 79,
        126,142,155, 137,154,166,  98,179,243, 110,190,254,
         70,215,255,  82,227,255,  42,252,255,  54,255,255,
#endif
    };

    uint32_t rgb_base[32];
    uint8_t *rgb = (uint8_t *)rgb_base;

    yuv2rgb((const uint8_t *[3]){Y+1,U+1,V+1}, (const int[3]){16,8,8}, rgb, 16,
            16, 2);
    for (int y = 0; y < 2; y++) {
        uint8_t *line = rgb + y*(16*4);
        for (int x = 0; x < 16; x++) {
            CHECK_PIXEL(&line[x*4], expect[(y*16+x)*3+0],
                        expect[(y*16+x)*3+1], expect[(y*16+x)*3+2], 255, x, y);
        }
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_asm_mips_conditions)
{
    static const uint8_t Y[1+32] =
        {0,50,60,70,80,90,100,110,120,130,140,150,160,170,180,190,200,
           55,65,75,85,95,105,115,125,135,145,155,165,175,185,195,205};
    static const ALIGNED(8) uint8_t U[1+8] =
        {0,  8, 40, 72,104,136,168,200,232};
    static const ALIGNED(8) uint8_t V[1+8] =
        {0,248,216,184,152,120, 88, 56, 24};
    static const uint8_t expect[32*3] = {
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        231,  0,  0, 230,  4,  0, 216, 22,  0, 202, 41,  0,
        188, 59,  0, 174, 78,  1, 161, 96, 45, 147,114, 89,
        133,133,133, 119,151,177, 105,170,221,  91,188,255,
         77,206,255,  63,225,255,  49,243,255,  48,255,255,
        237,  0,  0, 236, 10,  0, 222, 28,  0, 208, 47,  0,
        194, 65,  0, 180, 83,  7, 166,102, 51, 152,120, 95,
        139,139,139, 125,157,182, 111,175,226,  97,194,255,
         83,212,255,  69,230,255,  55,249,255,  54,255,255,
#else
        231,  0,  0, 243,  1,  0, 203, 26,  0, 215, 37,  0,
        176, 63,  0, 187, 74,  0, 148, 99, 61, 159,111, 73,
        120,136,149, 132,148,161,  92,173,237, 104,185,248,
         64,210,255,  76,221,255,  37,246,255,  48,255,255,
        237,  0,  0, 249,  7,  0, 209, 32,  0, 221, 43,  0,
        181, 68,  0, 193, 80,  0, 154,105, 67, 165,117, 79,
        126,142,155, 137,154,166,  98,179,243, 110,190,254,
         70,215,255,  82,227,255,  42,252,255,  54,255,255,
#endif
    };

    uint32_t rgb_base[32+1];
    uint8_t *rgb = (uint8_t *)rgb_base+1;

    yuv2rgb((const uint8_t *[3]){Y+1,U+1,V+1}, (const int[3]){16,8,8}, rgb, 16,
            16, 2);
    for (int y = 0; y < 2; y++) {
        uint8_t *line = rgb + y*(16*4);
        for (int x = 0; x < 16; x++) {
            CHECK_PIXEL(&line[x*4], expect[(y*16+x)*3+0],
                        expect[(y*16+x)*3+1], expect[(y*16+x)*3+2], 255, x, y);
        }
    }

    return 1;
}

#endif  // SIL_ARCH_MIPS

/*************************************************************************/
/**************************** Timing routines ****************************/
/*************************************************************************/

/* These routines are not "tests" per se, but this is a convenient place
 * to check the timing of the C vs. assembly routines. */

/*-----------------------------------------------------------------------*/

#if defined(SIL_ARCH_X86) && (defined(__GNUC__) || defined(__clang__))

TEST(time_x86)
{
    uint8_t *Y;
    ASSERT(Y = mem_alloc(640*480*3/2, 16, 0));
    uint8_t *U = Y + 640*480;
    uint8_t *V = U + 320*240;
    unsigned int seed = 1;
    for (int i = 0; i < 640*480*3/2; i++) {
        Y[i] = seed>>23 & 0xFF;
        seed = seed*22695477 + 1;
    }
    uint8_t *rgb;
    ASSERT(rgb = mem_alloc((640*480*4)+1, 16, 0));

    /* We make a first call to prime the CPU's caches and branch predictors;
     * then we measure 7 consecutive calls with the same arguments, discard
     * the lowest and highest times, and take the mean of the remainder.
     * These arrays are used to store the low-order word of the RDTSC value
     * at the beginning and end of each call. */
    uint32_t start_sse2[8], end_sse2[8];
    uint32_t start_c[8], end_c[8];

    /* We use a CPUID instruction to serialize execution so we measure only
     * the time taken by yuv2rgb() itself.  This modifies the EBX register,
     * which may be used as a PIC register on x86-32, so we define a macro
     * here to cover the architectural differences. */
    #ifdef SIL_ARCH_X86_32
        #define CPUID()  __asm__ volatile("push %%ebx; cpuid; pop %%ebx" \
                                          : : : "eax", "ecx", "edx")
    #else
        #define CPUID()  __asm__ volatile("cpuid" \
                                          : : : "eax", "ebx", "ecx", "edx")
    #endif

    for (int i = 0; i < 8; i++) {
        CPUID();
        __asm__ volatile("rdtsc" : "=a" (start_sse2[i]) : : "edx");
        yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){640,320,320},
                rgb, 640, 640, 480);
        CPUID();
        __asm__ volatile("rdtsc" : "=a" (end_sse2[i]) : : "edx");
    }

    for (int i = 0; i < 8; i++) {
        CPUID();
        __asm__ volatile("rdtsc" : "=a" (start_c[i]) : : "edx");
        yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){640,320,320},
                rgb+1, 640, 640, 480);
        CPUID();
        __asm__ volatile("rdtsc" : "=a" (end_c[i]) : : "edx");
    }

    #undef CPUID

    uint32_t time[7], low, high, total;

    for (int i = 0; i < 7; i++) {
        time[i] = end_sse2[i+1] - start_sse2[i+1];
        if (i == 0) {
            low = high = total = time[i];
        } else {
            total += time[i];
            if (time[i] < low) {
                low = time[i];
            } else if (time[i] > high) {
                high = time[i];
            }
        }
    }
    DLOG("SSE2: %u clocks (mean of 5/7)", (total+2)/5);

    for (int i = 0; i < 7; i++) {
        time[i] = end_c[i+1] - start_c[i+1];
        if (i == 0) {
            low = high = total = time[i];
        } else {
            total += time[i];
            if (time[i] < low) {
                low = time[i];
            } else if (time[i] > high) {
                high = time[i];
            }
        }
    }
    DLOG("C: %u clocks (mean of 5/7)", (total+2)/5);

    mem_free(Y);
    mem_free(rgb);
    return 1;
}

#endif  // SIL_ARCH_X86

/*-----------------------------------------------------------------------*/

/* This test could run on any MIPS platform, but MIPS doesn't expose a
 * timing register like RDTSC at the user privilege level, so we need
 * platform-specific help which we can only get on the PSP. */
#ifdef SIL_PLATFORM_PSP

TEST(time_mips_psp)
{
    uint8_t *Y;
    ASSERT(Y = mem_alloc(640*480*3/2, 16, 0));
    uint8_t *U = Y + 640*480;
    uint8_t *V = U + 320*240;
    unsigned int seed = 1;
    for (int i = 0; i < 640*480*3/2; i++) {
        Y[i] = seed>>23 & 0xFF;
        seed = seed*22695477 + 1;
    }
    uint8_t *rgb;
    ASSERT(rgb = mem_alloc((640*480*4)+1, 4, 0));

    uint32_t start, time;

    /* Run the routine once to prime the instruction cache, then again to
     * actually time it.  Since we disable interrupts, we can get a good
     * result from a single run. */
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){640,320,320},
            rgb, 640, 640, 480);
    psp_threads_lock();
    start = sceKernelGetSystemTimeLow();
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){640,320,320},
            rgb, 640, 640, 480);
    time = sceKernelGetSystemTimeLow() - start;
    psp_threads_unlock();
    DLOG("Assembly: %u usec", time);

    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){640,320,320},
            rgb+1, 640, 640, 480);
    psp_threads_lock();
    start = sceKernelGetSystemTimeLow();
    yuv2rgb((const uint8_t *[3]){Y,U,V}, (const int[3]){640,320,320},
            rgb+1, 640, 640, 480);
    time = sceKernelGetSystemTimeLow() - start;
    psp_threads_unlock();
    DLOG("C: %u usec", time);

    mem_free(Y);
    mem_free(rgb);
    return 1;
}

#endif  // SIL_PLATFORM_PSP

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
